 /*
 * dvdplayer.c: DVD player for DVBs PCI card
 *
 * Copyright (C) 2000 Matjaz Thaler
 *
 * CONTRIBUTORS:
 *  David Toth <dave@IRIS.geobio.elte.hu> 18-7-2000
 *        Communication between the player and the subtitle child 
 *        process.
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 * 
 * The author can be reached at matjaz.thaler@rd.iskraemeco.si 
 *
 * the project's page is at http://www2.arnes.si/~mthale1/index.html
 */


#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/types.h>		// semaphore 
#include <sys/ipc.h>		// semaphore, shared_memory
#include <sys/sem.h>		// semaphore
#include <sys/shm.h>		// shared memory
#include <sys/wait.h>		// waitpid
#include <signal.h>		// sigaction
#include <string.h>
#include <sys/poll.h>

#include <ost/video.h>
#include <ost/audio.h>
#include "css-descramble.h"
#include "css-init.h"
#include "subtitles.h"
#include "remote.h"
#include "dvdplayer.h"


void usage();
void synchronize(int file, unsigned char *buffer);
int setup_sem_shm_SIGINT();
void data_to_child(unsigned char*, int);
void interrupt_handler();
void change_file(char *fileName, int *file);
int next_file(char *file_name, int *file, int next_file);
int file_seek(int videoDev, int audioDev, char *file_name, int *file, int *go_to);
void exit_(int status);
int SendPCM(int size);

unsigned char title_key[5];
unsigned char diskKey[32]; 
unsigned char titleKey[32];
unsigned char sector[2048];
int inputFile=0;        // stdin
int osd=-1;
int video=-1;
int audio=-1;
struct pollfd pollfd_video;
struct pollfd pollfd_audio;

int semId=-1;		/* semaphore ID for subtitle child process */
struct data_buf *data_buffer;
int subtitlepid=-1;
int remotepid=-1;
char devVideo[128];
char devAudio[128];
char devOsd[128];
char devOsdOld[128];
FILE *ac3pipe;
int ac3_datasize=2048;


int main(int argc, char **argv){
  int c, i;
  int lpcm_count=0;

  if(setup_sem_shm_SIGINT()<0)
    exit_(-1);
  data_buffer->video_mode=V_PLAY;
  data_buffer->css=-1;
  data_buffer->ac3=-1;
  data_buffer->ac3stat=AC3_PLAY;
  data_buffer->lpcm=-1;
  data_buffer->mpeg_audio=-1;
  data_buffer->skip=0;
  data_buffer->subtitles=-1;
  data_buffer->subtitleStat=SUBTITLE_DISPLAY;
  data_buffer->remote=0;
  data_buffer->changeFile=CURRENT_FILE;
  data_buffer->port=-1;
  strcpy(data_buffer->directory, "/video");
  strcpy(devVideo, "/dev/ost/video");
  strcpy(devAudio, "/dev/ost/audio");
  strcpy(devOsd, "/dev/ost/osd");
  strcpy(devOsdOld, "/dev/video");
  strcpy(diskKey, "disc-key");
  strcpy(titleKey, "title-key");

  strcpy(data_buffer->ac3player, "");
  data_buffer->output_pointer_ac3 = 0;
  data_buffer->input_pointer_ac3 = 0;
  
  while(1){
    c = getopt(argc, argv, "va:l:mcf:g:s:rd:p:o:k:t:h::");
    if (c == -1)
      break;
    switch (c){
    case 'v':
      data_buffer->video_mode=V_OFF;
      fprintf (stderr, "do not playback video \n");
      break;
    case 'a':
      data_buffer->lpcm=-1; 
      if(optarg[0]=='a'){
	data_buffer->ac3=-2;
	fprintf (stderr, "ac3 sound\n");
      }
      else{
	sscanf(optarg, "%d", &data_buffer->ac3);
	data_buffer->ac3 &= 7;
	fprintf (stderr, "ac3 soundtrack %d\n", data_buffer->ac3);
      }
      break;
    case 'l':
      data_buffer->ac3=-1;
      if(optarg[0]=='a'){
	data_buffer->lpcm=-2;
	fprintf (stderr, "lpcm sound\n");
      }
      else{
	sscanf(optarg, "%d", &data_buffer->lpcm);
	data_buffer->lpcm &= 7;
	fprintf (stderr, "lpcm soundtrack %d\n", data_buffer->lpcm);
      }
      break;
    case 'm':
      data_buffer->mpeg_audio=0;
      fprintf (stderr, "mpeg audio\n");
      break;
    case 'c':
      data_buffer->css=1;
      fprintf (stderr, "css decode\n");
      break;
    case 'f':
      strcpy(data_buffer->fileName, optarg);
      inputFile=-1;
      break;
    case 'g':
      sscanf(optarg, "%d", &data_buffer->skip);
      fprintf (stderr, "skip %d MB\n", data_buffer->skip);
      break;
    case 's':
      sscanf(optarg, "%d", &data_buffer->subtitles);
      data_buffer->subtitles &= 31;
      fprintf (stderr, "subtitles %d\n", data_buffer->subtitles);
      break;
    case 'r':
      data_buffer->remote=1;
      fprintf (stderr, "use LIRC\n");
      if(data_buffer->port >= 0){
	fprintf (stderr, "you can use only one option: LIRC or network connection\n");
	exit_(1);
      }
      break;
    case 'p':
      sscanf(optarg, "%d", &data_buffer->port);
      fprintf (stderr, "listening on port %d\n", data_buffer->port);
      if(data_buffer->remote == 1){
	fprintf (stderr, "you can use only one option: LIRC or network connection\n");
	exit_(1);
      }
      break;
    case 'd':
      sscanf(optarg, "%d", &i);
      if(i<4 && i>=0){
	sprintf(devVideo, "%s%d", devVideo, i); 
	sprintf(devAudio, "%s%d", devAudio, i); 
	sprintf(devOsd, "%s%d", devOsd, i); 
	sprintf(devOsdOld, "%s%d", devOsdOld, i); 
	fprintf (stderr, "devices: %s %s %s\n", devVideo, devAudio, devOsd);
      }
      else{
	fprintf (stderr, "wrong device number: %s, you can only set device number: 0, 1, 2, or 3\n", optarg);
	exit_(1);
      }
      break;
    case 'o':
      strcpy(data_buffer->ac3player, optarg);
      fprintf (stderr, "ac3 decoder: %s\n", data_buffer->ac3player);
      break;
    case 'k':
      strcpy(diskKey, optarg);
      fprintf (stderr, "disk-key: %s\n", diskKey);
      break;
    case 't':
      strcpy(titleKey, optarg);
      fprintf (stderr, "title-key: %s\n", diskKey);
      break;
    case 'h':
    default:
      usage();
      exit_(1);
    }
  }
  // init
  if(data_buffer->css==1){
    if(css_init(title_key, diskKey, titleKey))
      exit_(1);
  }

  fprintf (stderr, "opening %s\n", devVideo);
  video=open(devVideo, O_RDWR | O_NONBLOCK);
  if(video < 0){
    perror("VIDEO DEVICE");
    exit_(1);
  }
  pollfd_video.fd = video;
  pollfd_video.events = POLLOUT;
 
  fprintf (stderr, "opening %s\n", devAudio);
  audio=open(devAudio, O_RDWR | O_NONBLOCK);
  if(audio < 0){
    perror("AUDIO DEVICE");
    exit_(1);
  }
  pollfd_audio.fd = audio;
  pollfd_audio.events = POLLOUT;
  
  fprintf (stderr, "opening %s\n", devOsd);
  osd=open(devOsd, O_RDWR);
  if(osd < 0){
    perror("OSD DEVICE");
    fprintf (stderr, "opening old API OSD device %s\n", devOsdOld);
    osd=open(devOsdOld, O_RDWR);
    if(osd < 0){
      perror("OSD DEVICE");
      exit_(1);
    }
  }
  
  if(ioctl(video, VIDEO_SELECT_SOURCE, VIDEO_SOURCE_MEMORY) < 0){
    perror("VIDEO_SELECT_SOURCE");
    exit_(1);
  }
  if(ioctl(audio, AUDIO_SELECT_SOURCE, AUDIO_SOURCE_MEMORY) < 0){
    perror("AUDIO_SELECT_SOURCE");
    exit_(1);
  }
  
  if(ioctl(audio, AUDIO_PLAY) < 0){
    perror("AUDIO_PLAY");
    exit_(1);
  }
  if(ioctl(video, VIDEO_PLAY) < 0){
    perror("VIDEO_PLAY");
    exit_(1);
  }
  
  if(data_buffer->remote > 0 || data_buffer->port > 0){
    if(!(remotepid=fork())){
      remote_control(video, audio, osd);
      exit_(0);
    }
    if(remotepid<0){
      fprintf(stderr,"ERROR: fork (remote child process) failed!\n");
      exit_(-1);
    }
  }
  if(data_buffer->ac3 != -1 || data_buffer->remote==1 || data_buffer->port>0){
    ac3dec_init();
  }

  if(data_buffer->subtitles >= 0 || data_buffer->remote==1 || data_buffer->port>0){
    if(!(subtitlepid=fork())){
      subtitles_decode(osd);
      exit_(0);
    }
    if(subtitlepid<0){
      fprintf(stderr,"ERROR: fork (subtitle child process) failed!\n");
      exit_(-1);
    }
  }
  if(data_buffer->ac3player != NULL){
    ac3pipe=popen(data_buffer->ac3player, "w");
    if(ac3pipe==NULL){
      fprintf (stderr, "can't run %s\n", data_buffer->ac3player);
      exit_(-1);
    }
  }
  if(inputFile!=0){
    inputFile=open(data_buffer->fileName, O_RDONLY);
    if(inputFile<0){
      fprintf (stderr, "can't open %s\n", data_buffer->fileName);
      exit_(1);
    }
    fprintf (stderr, "input file: %s\n", data_buffer->fileName);
  }
  if(data_buffer->skip && inputFile)
    file_seek(video, audio, data_buffer->fileName, &inputFile, &data_buffer->skip);

  while(1){                                                    //playback
    if(data_buffer->changeFile != CURRENT_FILE){
      if(data_buffer->changeFile == NEW_FILE)
	change_file(data_buffer->fileName, &inputFile);
      else{
	if(!next_file(data_buffer->fileName, &inputFile, data_buffer->changeFile)){
	  data_buffer->changeFile=CURRENT_FILE;
	  continue;
	}
      }
      data_buffer->changeFile=CURRENT_FILE;
      data_buffer->ac3stat=AC3_START;
      data_buffer->output_pointer_ac3 = data_buffer->input_pointer_ac3;
      ioctl(audio, AUDIO_CLEAR_BUFFER);
      ioctl(video, VIDEO_CLEAR_BUFFER);
    }
    if(data_buffer->go_to != 0){
      file_seek(video, audio, data_buffer->fileName, &inputFile, &data_buffer->go_to);
      data_buffer->ac3stat=AC3_START;
      data_buffer->output_pointer_ac3 = data_buffer->input_pointer_ac3;
      ioctl(audio, AUDIO_CLEAR_BUFFER);
      ioctl(video, VIDEO_CLEAR_BUFFER);
    }
    if(read(inputFile, sector, 2048) == 2048){                 //read input file
      if (!(sector[0]==0x00 && sector[1]==0x00 && sector[2]==0x01 && sector[3]==0xba)) {
	synchronize(inputFile, sector);
      }
      if ((sector[14+3]&0xFF)==0xE0){                                       // video
	if(data_buffer->video_mode >= V_PLAY){
	  if ((sector[20] & 0x30) && data_buffer->css==1) {                 // css decode
	    css_descramble(sector, title_key);
	    sector[20] &= 0x8f;
	  }
	  poll(&pollfd_video, 1, 100);
	  write(video, sector+14 , sector[19]+(sector[18]<<8)+6);
	  
	  // write LPCM audio
	  if(data_buffer->video_mode==V_PLAY){
	    ac3_datasize = (AC3_BUFFER_SIZE + data_buffer->input_pointer_ac3 - 
			    data_buffer->output_pointer_ac3) % AC3_BUFFER_SIZE;
	    if(ac3_datasize){
	      if(ac3_datasize > 1024*48)
		SendPCM(3096);
	      else if(ac3_datasize > 1024*32)
		SendPCM(1536);
	      else if((ac3_datasize > 1024*16) && !(lpcm_count%2))
		SendPCM(1536);
	      else if(ac3_datasize && !(lpcm_count%4))
		SendPCM(1536);
	      lpcm_count++;
	    }
	    else
	      lpcm_count=0;	    
	  }
	}
      }
      else if (((sector[14+3]&0xF0)==0xC0) && (data_buffer->mpeg_audio==0)){    // mpeg audio
	if ((sector[20] & 0x30) && data_buffer->css==1) {                       // css decode
	  css_descramble(sector, title_key);
	  sector[20] &= 0x8f;
	}
	poll(&pollfd_audio, 1, 200);
	write(audio, sector+14 , sector[19]+(sector[18]<<8)+6);
      }
      else if (((data_buffer->ac3>=0) && ((sector[14+sector[14+8]+9])==(0x80|data_buffer->ac3))) || 
	       ((data_buffer->ac3==-2) && ((sector[14+sector[14+8]+9]&0xF8)==0x80))) {  // Dolby AC3 audio
	if ((sector[20] & 0x30) && data_buffer->css==1) {                 // css decode
	  css_descramble(sector, title_key);
	  sector[20] &= 0x8f;
	}

	if(strcmp(data_buffer->ac3player, "") && (data_buffer->ac3stat != AC3_STOP))
	  fwrite(sector+sector[22]+27, ((sector[18]<<8)|sector[19])-sector[22]-7, 1, ac3pipe);
	else if(data_buffer->ac3stat == AC3_PLAY)
	  ac3dec_decode_data(sector+sector[22]+27, sector+20+((sector[18]<<8)|sector[19]), 0, &data_buffer->input_pointer_ac3, &data_buffer->output_pointer_ac3, data_buffer->ac3_data);
	else if(data_buffer->ac3stat == AC3_START){
	  ac3dec_decode_data(sector+sector[22]+27, sector+20+((sector[18]<<8)|sector[19]), 1, &data_buffer->input_pointer_ac3, &data_buffer->output_pointer_ac3, data_buffer->ac3_data);
	  data_buffer->ac3stat = AC3_PLAY;
	}
	if((data_buffer->video_mode) < V_PLAY && (data_buffer->ac3stat != AC3_STOP)){
	  SendPCM(2048*16);
	}
      }
      else if (((data_buffer->lpcm>=0) && ((sector[14+sector[14+8]+9])==(0xA0|data_buffer->lpcm))) || 
	       ((data_buffer->lpcm==-2) && ((sector[14+sector[14+8]+9]&0xF8)==0xA0))) {  // LPCM audio
	if ((sector[20] & 0x30) && data_buffer->css==1) {                 // css decode
	  css_descramble(sector, title_key);
	  sector[20] &= 0x8f;
	}
	poll(&pollfd_audio, 1, 200);
	write(audio, sector+14 , sector[19]+(sector[18]<<8)+6);
      }
      else if ((data_buffer->subtitles>=0) && (sector[14+sector[14+8]+9])==
	       (0x20|data_buffer->subtitles)) {                           //spu data
	if ((sector[20] & 0x30) && data_buffer->css==1) {                 // css decode
	  css_descramble(sector, title_key);
	  sector[20] &= 0x8f;
	}
	i = 24 + sector[22];
	data_to_child(sector+i , 2048-i);
      }
    }
    else{                                                   // end of file
      if(inputFile!=0){
	if(next_file(data_buffer->fileName, &inputFile, 1))
	  continue;
	close(inputFile);
      }
      if(data_buffer->remote || data_buffer->port>0){                              // select new file
	while(1){
	  if(data_buffer->exit_child)
	    break;
	  sleep(1);
	  if(data_buffer->changeFile != CURRENT_FILE){
	    inputFile=open(data_buffer->fileName, O_RDONLY);              
	    if(inputFile<0){
	      fprintf (stderr, "can't open %s\n", data_buffer->fileName);
	      data_buffer->changeFile=CURRENT_FILE;
	    }
	    else{
	      fprintf (stderr, "playing: %s\n", data_buffer->fileName);
	      break;
	    }
	  }
	}
      }
      else
	break;
    }
    if(data_buffer->exit_child) 
      break;
  }
  exit_(0);
  return 0;
} 

void usage(){
  fprintf (stderr, "\n");
  fprintf (stderr, "Usage:\tdvdplayer [OPTION]...\n");
  fprintf (stderr, "OPTIONS:\n");
  fprintf (stderr, " -f filename\topen file [stdin]\n");
  fprintf (stderr, " -c\t\tcss decode (playback from DVD)\n");
  fprintf (stderr, " -a channel\tplay ac3 sound\n");
  fprintf (stderr, " -l channel\tplay lpcm sound\n");
  fprintf (stderr, " -m \t\tplay mpeg sound\n");
  fprintf (stderr, " -s subtitle  \tdisplay subtitles\n");
  fprintf (stderr, " -v\t\tdo not play video\n");
  fprintf (stderr, " -g n\t\tstart playback at n M bytes of input file\n");
  fprintf (stderr, " -r \t\tremote control with LIRC\n");
  fprintf (stderr, " -p port\tcontrol via a network connection\n");
  fprintf (stderr, " -d dev\t\tdevice number (0, 1, 2, 3)   [/dev/ost/video ...]\n");
  fprintf (stderr, " -o decoder\tsend ac3 stream to decoder (ac3play, ac3spdif)\n");
  fprintf (stderr, " -k filename\tdisk key  [disk-key]\n");
  fprintf (stderr, " -t filename\ttitle key [title-key]\n");
  fprintf (stderr, " -h\t\tshow this usage information\n");
  fprintf (stderr, "\n");
  fprintf (stderr, "EXAMPLE:\n");
  fprintf (stderr, " -playback from DVD:\n");
  fprintf (stderr, "   dvdplayer -a 0 -c -s 0 -f /dvd/video_ts/vts_01_1.vob\n");
  fprintf (stderr, " -playback from hard disk:\n");
  fprintf (stderr, "   dvdplayer -a 0 -s 0 -f foo.vob\n");
  fprintf (stderr, "\n");
}

void synchronize(int file, unsigned char *buffer){

  int sync=0;
  
  fprintf(stderr, "sync\n");
  while(sync==0){
    while(buffer[0] != 0)
      read(file, &buffer[0], 1);
    read(file, &buffer[1], 1);
    if(buffer[1] != 0)
      continue;
    buffer[2]=0;
    while(buffer[2] == 0)
      read(file, &buffer[2], 1);
    if(buffer[2] != 1)
      continue;
    read(file, &buffer[3], 1);
    if(buffer[3] != 0xba)
      continue;
    read(file, &buffer[4], 2044);
    sync=1;
  }
}

int setup_sem_shm_SIGINT()
{
  union semun arg;
  int shmid_buf;
  struct sigaction new_sa, old_sa;

  // setup a new interrupt handler
  new_sa.sa_handler = interrupt_handler;
  sigemptyset(&new_sa.sa_mask);
  new_sa.sa_flags = 0;
  if (sigaction(SIGINT, &new_sa, &old_sa) == -1) 
    return (-1);

  // create a semaphore
  if((semId = semget(IPC_PRIVATE, 2, 0600)) < 0){
    fprintf(stderr,"ERROR: semget failed, errno is"); perror(" ");
    return(-1);
  }

  arg.val=0;
  if(semctl(semId, 0, SETVAL, arg) == -1){
    semctl(semId, 1, IPC_RMID, arg);	// remove semaphore
    semId=-1;
    fprintf(stderr,"ERROR: semctl failed, errno is"); perror(" ");
    return(-1);
  }

  // create shared memory 
  if((shmid_buf = shmget(IPC_PRIVATE, sizeof(struct data_buf),
                         IPC_CREAT|0600)) == -1){
    semctl(semId, 1, IPC_RMID, arg);	/* remove semaphore */
    semId=-1;
    fprintf(stderr,"ERROR: shmget failed, errno is"); perror(" ");
  }
  data_buffer = (struct data_buf*) shmat(shmid_buf, 0, 0);
  shmctl(shmid_buf, IPC_RMID, 0);
  
  // init buffers
  data_buffer->exit_child=0;
  data_buffer->input_pointer=0;
  data_buffer->output_pointer=0;
  data_buffer->input_pointer_ac3=0;
  data_buffer->output_pointer_ac3=0;
  return(0);
}

void data_to_child(unsigned char *data, int datasize)
{
  while(datasize){
    if( ((data_buffer->input_pointer+1)%SUBTITLE_BUFFER_SIZE) !=
	data_buffer->output_pointer){    // There is room in the buffer
      data_buffer->data[data_buffer->input_pointer]=*data++;
      datasize--;
      data_buffer->input_pointer=
	(data_buffer->input_pointer+1)%SUBTITLE_BUFFER_SIZE;
    }
    else{                           // Buffer is full
      if(sem_ctl(semId, 0, 0))
	return;
      if(sem_op(semId, 0, -1))      // sleep until child is done
	return;
    }
  }
  if(sem_ctl(semId, 0, 0))          // wake up child if sleeps
    return;
}

void interrupt_handler()
{
  if(semId>=0) semctl(semId, 1, IPC_RMID, NULL);	/* remove semaphore */
  semId=-1;
  if(!data_buffer->exit_child) 
    fprintf(stderr,"Interrupt signal received.\n");
  data_buffer->exit_child=1;
}

void change_file(char *fileName, int *file){

  int f;

  f=open(fileName, O_RDONLY);
  if(f<0)
    fprintf (stderr, "can't open %s\n", fileName);
  else{
    close(*file);
    *file=f;
  }
}

int next_file(char *file_name, int *file, int next_file){
  
  int len=strlen(file_name);
  int f;
  
  if(next_file){
    if(!(file_name[len-5]>='1' && file_name[len-5]<='9' && len>=12))
      return 0;
    file_name[len-5]+=next_file;
    if(!(file_name[len-5]>='1' && file_name[len-5]<='9' && len>=12)){
      file_name[len-5]-=next_file;
      return 0;
    }
    f=open(file_name, O_RDONLY);
    if(f<0){
      fprintf (stderr, "can't open %s\n", file_name);
      file_name[len-5]-=next_file;
      return 0;
    }
    else{
      fprintf (stderr, "opening %s\n", file_name);
      close(*file);
      *file=f;
      return 1;
    }
  }
  return 0;
}

int file_seek(int videoDev, int audioDev, char *file_name, int *file, int *go_to){

  //  struct video_play_mode pmode;
  long long int i, end, cur, seek;

  seek=*go_to;
  seek*=1024*1024;
  cur=lseek(*file, 0, SEEK_CUR);
  end=lseek(*file, 0, SEEK_END);
  while(seek){
    i=cur+seek;
    if(!(i<0 || i>end)){
      lseek(*file, i, SEEK_SET);
      seek=0;
    }
    else if(i<0){
      if(next_file(file_name, file, -1)){
	seek+=cur;
	cur=end=lseek(*file, 0, SEEK_END);
      }
      else{
	lseek(*file, 0, SEEK_SET);
	break;
      }
    }
    else{
      if(next_file(file_name, file, 1)){
	seek-=end-cur;
	cur=0;
	end=lseek(*file, 0, SEEK_END);
      }
      else{
	lseek(*file, -4*1024*1024, SEEK_END);
	break;
      }
    }
  }
  *go_to=0;
  if(seek==0)
    return 1;
  else
    return 0;
}

int sem_ctl(int semid, int semnum, int val){

  union semun arg;

  arg.val=val;	
  if(semctl(semid, semnum, SETVAL, arg) == -1){
    if(!data_buffer->exit_child){
      fprintf(stderr,"ERROR: semctl failed, errno is"); perror(" ");
      data_buffer->exit_child=1;
    }
    return 1;
  }
  return 0;
}

int sem_op(int semid, int semnum, int op){

  struct sembuf semBuf;

  semBuf.sem_num = semnum;
  semBuf.sem_op  = op; 
  semBuf.sem_flg = 0;
  if(semop(semid,&semBuf,1)){
    if(!data_buffer->exit_child){
      fprintf(stderr,"ERROR: semop failed, errno is"); perror(" ");
      data_buffer->exit_child=1;
    }
    return 1;
  }
  return 0;
}

void exit_(int status){

  data_buffer->exit_child=1;
  usleep(100000);
  if(subtitlepid > 0)
    kill(subtitlepid, 15);
  if(remotepid > 0)
    kill(remotepid, 15);
  if(video > 0)
    close(video);
  if(audio > 0)
    close(audio);
  if(osd > 0)
    close(osd);
  if(inputFile > 0)
    close(inputFile);
  if(strcmp(data_buffer->ac3player, ""))
    pclose(ac3pipe);
  if(semId >= 0) 
    semctl(semId, 1, IPC_RMID, NULL);	  //remove semaphore
  exit(status);
}


// data=PCM samples, 16 bit, LSB first, 48kHz, stereo
int SendPCM(int size){

#define MAXSIZE 2032

  char buffer[MAXSIZE+16];
  int length=0;
  int p_size;

  if(data_buffer->output_pointer_ac3 == data_buffer->input_pointer_ac3)
    return 1;

  while (size > 0) {
    if(size >= MAXSIZE)
      p_size=MAXSIZE;
    else
      p_size=size;
    length=10;

    while(p_size){
      if(data_buffer->output_pointer_ac3 != data_buffer->input_pointer_ac3){ // data in the buffer
	buffer[(length+6) ^ 1]=data_buffer->ac3_data[data_buffer->output_pointer_ac3]; // swab because ac3dec delivers wrong byteorder
	p_size--;
	length++;
	data_buffer->output_pointer_ac3 = (data_buffer->output_pointer_ac3+1)%AC3_BUFFER_SIZE;
      }
      else
	break;
    }

    buffer[0] = 0x00;
    buffer[1] = 0x00;
    buffer[2] = 0x01;
    buffer[3] = 0xbd;
    
    buffer[4] = (length >> 8) & 0xff;
    buffer[5] = length & 0xff;
    
    buffer[6] = 0x80;
    buffer[7] = 0x00;
    buffer[8] = 0x00;
    
    buffer[9] = 0xa0;  // substream ID
    buffer[10] = 0x00; // other stuff (see DVD specs), ignored by driver
    buffer[11] = 0x00;
    buffer[12] = 0x00;
    buffer[13] = 0x00;
    buffer[14] = 0x00;
    buffer[15] = 0x00;

    poll(&pollfd_audio, 1, 200);
    write(audio, buffer, length+6);
    size-=MAXSIZE;
  }
  return 0;
}
